"""
Shared constants for Biblicus.
"""

SCHEMA_VERSION = 2
DATASET_SCHEMA_VERSION = 1
CORPUS_DIR_NAME = ".biblicus"
DEFAULT_RAW_DIR = "raw"
SIDECAR_SUFFIX = ".biblicus.yml"
RUNS_DIR_NAME = "runs"
EXTRACTION_RUNS_DIR_NAME = "extraction"
HOOK_LOGS_DIR_NAME = "hook_logs"
